/*
 *  DataFromFiles.cpp
 *  LogitReg
 *
 *  Created by Youna Hu on 1/14/11.
 *  Copyright 2011 U of Michigan. All rights reserved.
 *
 */

#include "DataFromFiles.h"

void DataFromFiles::Initializer(){
	colSize = 0;
	fileStep = 0;
	header = false;
	IDcolumns = false;
}

void DataFromFiles::Initializer(bool headerIN, bool IDcolumnsIn){
	header = headerIN;
	IDcolumns = IDcolumnsIn;
	colSize = 0;
	fileStep = 0;
}

DataFromFiles::DataFromFiles(){
	Initializer();
}

DataFromFiles::DataFromFiles(bool headerIN, bool IDcolumnsIn){
	Initializer(headerIN, IDcolumnsIn);
}

void DataFromFiles::IDReadAsString(String &indvIDFileName){
	indvsIDs.clear();
	FILE *IDFile = fopen(indvIDFileName, "r");
	if (IDFile == NULL){
		error("Error in opening file (%s)\n",indvIDFileName.c_str());
	}
	// read in the individual file
	fileStep = 0;
	while (!feof(IDFile)) {
		fileEntry.ReadLine(IDFile);
		if (fileEntry.IsEmpty()) {
			error("Error: File %s has an empty line (%d) \n",indvIDFileName.c_str(),fileStep);
		}
		fileToken.AddTokens(fileEntry, "\t");
		if (fileToken.Length() != 1) {
			error("Error: Wrong format of the 012.indv file %s", indvIDFileName.c_str());
		}
		// store the individual IDs 
		if (header) { // if the header is true, then store the IDs starting from the second line, since the first is a name
			if (fileStep >= 1) {
				indvsIDs.push_back(fileToken[0]);
			}
		} else { // else, store the IDs starting from the first line
			indvsIDs.push_back(fileToken[0]);
		}
		
		fileEntry.Clear();
		fileToken.Clear();
		fileStep ++;
	}	
	fclose(IDFile);
}

void DataFromFiles::File2Vector(String &inFileName, Vector &vecOut){ 
	// Use the implementation for File2Matrix and then translate it into a vector
	File2Matrix(inFileName, x);
	nSample = x.rows;
	colSize = x.cols;
	if (colSize != 1) {
		error("Error in reading %s into a matrix.\n",inFileName.c_str());
	}
	for (int i = 0; i < nSample; i ++) {
		vecOut[i] = x[i][0];
	}
}

void DataFromFiles::File2Vector(String &inFileName, String &indvIDFileName, Vector &vecOut){
	// phenoValue contains ID + value column
	map_IDAndValue phenoValue;
	FILE *phenoMatFile = fopen(inFileName, "r");
	if (phenoMatFile == NULL){
		error("Error in opening file (%s))\n",inFileName.c_str());
	}
	
	// read in the ID as String vectors
	IDReadAsString(indvIDFileName);
	fileStep = 0;
	
	// read in the file with ID
	while (!feof(phenoMatFile)) {
		fileEntry.ReadLine(phenoMatFile);
		if (fileEntry.IsEmpty()) {
			error("Error: File %s has an empty line (%d) \n",inFileName.c_str(),fileStep);
		}
		fileToken.AddTokens(fileEntry, "\t");
		
		if (fileToken.Length() != 2) {
			error("Error: Wrong format of the 012.indv file %s", inFileName.c_str());
		} 
		
		// store the phenotype data into the map
		if (header) {
			if (fileStep >=1 ) {
				phenoValue.insert(map_IDAndValue::value_type(fileToken[0],fileToken[1].AsInteger()));
			}
		} else {
			phenoValue.insert(map_IDAndValue::value_type(fileToken[0],fileToken[1].AsInteger()));
		}

		fileEntry.Clear();
		fileToken.Clear();
		fileStep ++;
	}
	fclose(phenoMatFile);
	
	nSample = indvsIDs.size();
	vecOut.Dimension(nSample);
	// update the vector y according to the order of indvIDs
	for (int i = 0; i < nSample; i ++) {
		vecOut[i] = phenoValue[indvsIDs[i]];
	}	
	
}

void DataFromFiles::File2Matrix(String &covFileName, Matrix &matOut){
	FILE *genoMat = fopen(covFileName, "r");
	
	if (genoMat == NULL) {
		printf("Error opening file ( %s )!", covFileName.c_str());
		return;
	}
	
	fileStep = 0;
	while (!feof(genoMat)) {
		// the line and tokens 
		fileEntry.ReadLine(genoMat);
		if(fileEntry.IsEmpty()) continue;
		fileToken.AddTokens(fileEntry, "\t");
		colSize = fileToken.Length();
		if (header) { // if the first line is header line, do no read in the first line into the matrix
			if (fileStep >= 1) {
				if (fileStep == 0){
					xLine.Dimension(1, colSize);
					matOut.Dimension(1, colSize);
					xLine.Zero();
					matOut.Zero();
				}		
				for (int i = 0; i < colSize; i ++) {
					xLine[0][i] = fileToken[i];
				} 
				if (fileStep == 0){
					matOut = xLine;
				} else {
					matOut.StackBottom(xLine);
				}							
			}
		} else { // if the first line is not header line, reading starts with the first line
			if (fileStep == 0){
				xLine.Dimension(1, colSize);
				matOut.Dimension(1, colSize);
				xLine.Zero();
				matOut.Zero();
			}		
			for (int i = 0; i < colSize; i ++) {
				xLine[0][i] = fileToken[i];
			} 
			if (fileStep == 0){
				matOut = xLine;
			} else {
				matOut.StackBottom(xLine);
			}			
		}
		
		fileStep ++;
		fileToken.Clear();
		fileEntry.Clear();
	}
	fclose(genoMat);
}

void DataFromFiles::File2Matrix(String &covFileName, String &indvIDFileName, Matrix &matOut){
	// read in the ID as string vectors
	IDReadAsString(indvIDFileName);

	FILE *covMatFile = fopen(covFileName, "r");
	
	if (covMatFile == NULL){
		printf("Error in opening file (%s)\n", covFileName.c_str());
	}
	
	fileStep = 0;
	map_IDAndValue IDCovIdex; // store ID and the index/order, 1,2,3,...
	
	// read in covFile as matrix -- this is different from phenoFileAlign which read in a phenotype vector
	while (!feof(covMatFile)) {
		fileEntry.ReadLine(covMatFile);
		if (fileEntry.IsEmpty()) {
			error("Error: File %s has an empty line (%d) \n", covFileName.c_str(),fileStep);
		}
		fileToken.AddTokens(fileEntry, "\t");
		colSize = fileToken.Length();
		// store the ID + fileStep, genotype matrix separately 
		// for now, store the genotype matrix into x, and then reorder it
		if (header) {
			if (fileStep >= 1) { // if the first line is a header, then ignore the first line of the file
				IDCovIdex.insert(map_IDAndValue::value_type(fileToken[0],fileStep));
				if (fileStep == 1) {
					xLine.Dimension(1,colSize-1,0);
					x.Dimension(1,colSize-1,0);
				}
				
				for (int i = 1; i < colSize; i ++) {
					xLine[0][i-1] = fileToken[i];
				}
				
				if (fileStep == 1) {
					x = xLine;
				} else {
					x.StackBottom(xLine);
				}				
			}
		} else {
			IDCovIdex.insert(map_IDAndValue::value_type(fileToken[0],fileStep));
			if (fileStep == 0) {
				xLine.Dimension(1,colSize-1,0);
				x.Dimension(1,colSize-1,0);
			}
			
			for (int i = 1; i < colSize; i ++) {
				xLine[0][i-1] = fileToken[i];
			}
			if (fileStep == 0) {
				x = xLine;
			} else {
				x.StackBottom(xLine);
			}				
		}		
		fileStep ++;
		fileEntry.Clear();
		fileToken.Clear();				
	}
	fclose(covMatFile);
	
	nSample = indvsIDs.size();
	matOut.Dimension(x.rows, x.cols);
	// when write out the matOut matrix, make sure that the ID matches with indvsIDs read from the .012.indv file
	for (int i = 0; i < nSample; i++) {
		for (int j = 0; j < colSize -1; j++) {
			// ID = indvSIDs[i] make sure the order is given by the .012.indv file
			// IDCovIdex[ID] looks for the correponding index of ID in the file covFileName 
			matOut[i][j] = x[IDCovIdex[indvsIDs[i]]][j];
		}
		//	printf("new key = %d\n",IDCovIdex[indvsIDs[i]]);
	}
}
void DataFromFiles::PositionFile2Matrix(String &covFileName, Matrix &matOut){
	FILE *genoMat = fopen(covFileName, "r");
	
	if (genoMat == NULL) {
		printf("Error opening file ( %s )!", covFileName.c_str());
		return;
	}
	
	fileStep = 0;
	while (!feof(genoMat)) {
		// the line and tokens 
		fileEntry.ReadLine(genoMat);
		if(fileEntry.IsEmpty()) continue;
		fileToken.AddTokens(fileEntry, "\t");
		colSize = fileToken.Length();
		if (header) { // if the first line is header line, do not read in the first line into the matrix
			if (fileStep >= 1) {
				if (fileStep == 0){
					xLine.Dimension(1, colSize);
					matOut.Dimension(1, colSize);
					xLine.Zero();
					matOut.Zero();
				}		
				for (int i = 0; i < colSize; i ++) {
					// take care of chromosome X/x or Y/y
					if (fileToken[i] == "X" || fileToken[i] == "x") {
						xLine[0][i] = 23;
					} else if (fileToken[i] == "Y" || fileToken[i] == "y") {
						xLine[0][i] = 24;
					} else {
						xLine[0][i] = fileToken[i];
					}
				} 
				if (fileStep == 0){
					matOut = xLine;
				} else {
					matOut.StackBottom(xLine);
				}							
			}
		} else { // if the first line is not header line, reading starts with the first line
			if (fileStep == 0){
				xLine.Dimension(1, colSize);
				matOut.Dimension(1, colSize);
				xLine.Zero();
				matOut.Zero();
			}		
			for (int i = 0; i < colSize; i ++) {
				// take care of chromosome X/x or Y/y
				if (fileToken[i] == "X" || fileToken[i] == "x") {
					xLine[0][i] = 23;
				} else if (fileToken[i] == "Y" || fileToken[i] == "y") {
					xLine[0][i] = 24;
				} else {
					xLine[0][i] = fileToken[i];
				}
			} 
			if (fileStep == 0){
				matOut = xLine;
			} else {
				matOut.StackBottom(xLine);
			}			
		}
		
		fileStep ++;
		fileToken.Clear();
		fileEntry.Clear();
	}
	fclose(genoMat);	
}

std::vector<String> & DataFromFiles::GetIndvsIDs(){
	return(indvsIDs);
}